// study area
var studyArea = ee.FeatureCollection('users/river extraction/LCJ');
Map.addLayer(studyArea, {color: "yellow"}, "Study Area", true);
Map.centerObject(studyArea, 7);

////////////////////////////////////////
// Composite time period
// // // Time period
var Filter1 = ee.Filter.date('2009-08-11','2009-08-27');  
var Filter2 = ee.Filter.date('2010-07-18','2010-07-23');
var Filter3 = ee.Filter.date('2007-09-04','2007-09-18');

// // composite date
var dateFilter = ee.Filter.or(Filter1, Filter2, Filter3);

////////////////////////////////////////
// Composite parameters
// cloudThresh: If using the cloudScoreTDOMShift method-Threshold for cloud 
//     masking (lower number masks more clouds.  Between 10 and 30 generally 
//     works best)
var cloudThresh = 20; 
   
// dilatePixels: Number of pixels to buffer clouds and cloud 
//     shadows by (1 or 2 generally is sufficient)
var dilatePixels = 1;    
	
// cloudHeights: Height of clouds to use to project cloud shadows
var cloudHeights = ee.List.sequence(200,5000,500);

// zScoreThresh: Threshold for cloud shadow masking- lower number masks out 
// less.  Between -0.8 and -1.2 generally works well
var zScoreThresh = -8000;    
	
// shadowSumThresh: Sum of IR bands to include as shadows within TDOM and the 
// shadow shift method (lower number masks out less)
var shadowSumThresh = 3000;

///////////////////////////////////////////////////////////////////////////////
//Function for acquiring Landsat TOA image collection
// function getImageCollection(studyArea,dateFilter){
function getImageCollection(studyArea,dateFilter){
  var ls;var l5SRs;var l7SRs;var l8SRs;var out;
  
  var sensorBandDictLandsatSR =ee.Dictionary({L8 : ee.List([1,2,3,4,5,9,6]),
                        L7 : ee.List([0,1,2,3,4,5,7]),
                        L5 : ee.List([0,1,2,3,4,5,6])
  });
  var bandNamesLandsatSR = ee.List(['blue','green','red','nir','swir1','temp',
      'swir2']);
  
  l5SRs = ee.ImageCollection('LANDSAT/LT05/C01/T1_SR')
      .filter(dateFilter)
      .filterBounds(studyArea)
      .select(sensorBandDictLandsatSR.get('L5'),bandNamesLandsatSR);
  
  l8TOAs = ee.ImageCollection('LANDSAT/LC08/C01/T1_TOA')
      .filter(dateFilter)
      .filterBounds(studyArea)
      .select(sensorBandDictLandsatSR.get('L8'),bandNamesLandsatSR);

  l7SRs = ee.ImageCollection('LANDSAT/LE07/C01/T1_SR')
      .filter(dateFilter)
      .filterBounds(studyArea)
      .select(sensorBandDictLandsatSR.get('L7'),bandNamesLandsatSR);
  ls = ee.ImageCollection(l5SRs.merge(l7SRs).merge(l8TOAs));
  out = ls;
  return out;
}

///////////////////////////////////////////////////////////////////////////////
// A helper to apply an expression and linearly rescale the output.
// Used in the landsatCloudScore function below.
var rescale = function(img, exp, thresholds) {
  return img.expression(exp, {img: img})
      .subtract(thresholds[0]).divide(thresholds[1] - thresholds[0]);
};

///////////////////////////////////////////////////////////////////////////////
// Compute a cloud score and adds a band that represents the cloud mask.  
// This expects the input image to have the common band names: 
// ["red", "blue", etc], so it can work across sensors.
function landsatCloudScore(img) {
  // Compute several indicators of cloudiness and take the minimum of them.
  var score = ee.Image(1.0);
  // Clouds are reasonably bright in the blue band.
  score = score.min(rescale(img, 'img.blue', [1000, 3000]));
 
  // Clouds are reasonably bright in all visible bands.
  score = score.min(rescale(img, 'img.red + img.green + img.blue', [2000, 8000]));
   
  // Clouds are reasonably bright in all infrared bands.
  score = score.min(
      rescale(img, 'img.nir + img.swir1 + img.swir2', [3000, 8000]));

  // Clouds are reasonably cool in temperature.
  score = score.where(img.select(['temp']).mask(),score.min(rescale(img,
      'img.temp', [3000, 2900])));

  // However, clouds are not snow.
  var ndsi = img.normalizedDifference(['green', 'swir1']);
  var inv_cloudiness = score.min(rescale(ndsi, 'img', [0.8, 0.6]));

  score =  score.min(rescale(ndsi, 'img', [0.8, 0.6])).multiply(100).byte();
  score = score.lt(cloudThresh).rename('cloudMask');
  img = img.updateMask(img.mask().and(score));
  return img.addBands(score);
}

///////////////////////////////////////////////////////////////////////////////
//Function for finding dark outliers in time series.
//Original concept written by Carson Stam and adapted by Ian Housman.
//Adds a band that is a mask of pixels that are dark, and dark outliers.
function simpleTDOM2(collection,zScoreThresh,shadowSumThresh,dilatePixels){
  var shadowSumBands = ['nir','swir1'];
  
  //Get some pixel-wise stats for the time series
  var irStdDev = collection.select(shadowSumBands).reduce(ee.Reducer.stdDev());
  var irMean = collection.select(shadowSumBands).mean();
  
  //Mask out dark dark outliers
  collection = collection.map(function(img){
    var zScore = img.select(shadowSumBands).subtract(irMean).divide(irStdDev);
    var irSum = img.select(shadowSumBands).reduce(ee.Reducer.sum());
    var TDOMMask = zScore.lt(zScoreThresh).reduce(ee.Reducer.sum()).eq(2)
        .and(irSum.lt(shadowSumThresh)).not();
    TDOMMask = TDOMMask.focal_min(dilatePixels);
    return img.addBands(TDOMMask.rename('TDOMMask'));
  });
  return collection;
}

///////////////////////////////////////////////////////////////////////////////
//Function for wrapping cloud and shadow masking together.
//Assumes image has cloud mask band called "cloudMask" and a TDOM mask called 
//"TDOMMask".
function cloudProject(img,shadowSumThresh,dilatePixels,cloudHeights){
  
  //Get the cloud mask
  var cloud = img.select('cloudMask').not();
  cloud = cloud.focal_max(dilatePixels);
  cloud = cloud.updateMask(cloud);
  
  //Get TDOM mask
  var TDOMMask = img.select(['TDOMMask']).not();
  
  //Project the shadow finding pixels inside the TDOM mask that are dark and 
  //inside the expected area given the solar geometry
  var darkPixels = img.select(['nir','swir1','swir2'])
      .reduce(ee.Reducer.sum()).gte(1);
  
  //Get scale of image
  var nominalScale = cloud.projection().nominalScale();

  //Find where cloud shadows should be based on solar geometry
  //Convert to radians
  var meanAzimuth = img.get('SUN_AZIMUTH');
  var meanZenith = img.get('SUN_ELEVATION');
  var azR = ee.Number(meanAzimuth).multiply(Math.PI).divide(180.0)
      .add(ee.Number(0.5).multiply(Math.PI ));
  var zenR = ee.Number(0.5).multiply(Math.PI )
      .subtract(ee.Number(meanZenith).multiply(Math.PI).divide(180.0));
  
  //Find the shadows
  var shadows = cloudHeights.map(function(cloudHeight){
    cloudHeight = ee.Number(cloudHeight);
    var shadowCastedDistance = zenR.tan()
        .multiply(cloudHeight);//Distance shadow is cast
    var x = azR.cos().multiply(shadowCastedDistance)
        .divide(nominalScale).round();//X distance of shadow
    var y = azR.sin().multiply(shadowCastedDistance)
        .divide(nominalScale).round();//Y distance of shadow
    return cloud.changeProj(cloud.projection(), cloud.projection()
        .translate(x, y));
  });

  var shadow = ee.ImageCollection.fromImages(shadows).max();

  //Create shadow mask
  shadow = shadow.updateMask(shadow.mask().and(cloud.mask().not()));
  shadow = shadow.focal_max(dilatePixels);
  shadow = shadow.updateMask(shadow.mask().and(darkPixels).and(TDOMMask));

  //Combine the cloud and shadow masks
  var combinedMask = cloud.mask().or(shadow.mask()).eq(0);
  
  //Update the image's mask and return the image
  img = img.updateMask(img.mask().and(combinedMask));
  img = img.addBands(combinedMask.rename(['cloudShadowMask']));
  return img;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// vizParams: Options for map visualization
var vizParams = {'min': 0.05,'max': 0.3, 'bands':'red,green,blue', 
    'gamma':1.6};
var vizParamsRescale = {'min': 500,'max': 3000, 'bands':'red,green,blue', 
    'gamma':1.6};
var vizParamsFalseColor = {'min':0, 'max': [5000, 3000, 3000], 'bands':'nir,red,green'};

var vis = {
  'Raw': {'min': 500,'max': 3000, 'bands':'red,green,blue', 'gamma':1.6},
  'RGB': {'min': 500,'max': 3000, 'bands':'red,green,blue', 'gamma':1.6},
  'False Color': {'min':0, 'max': [5000, 3000, 3000], 'bands':'nir,red,green'}
};

// Get Landsat 5,8,7 Image Collections
var ls = getImageCollection(studyArea,dateFilter);

var raw = ls.median().clip(studyArea)
Map.addLayer(ls.median().clip(studyArea), vizParams,
  'Landsat before masking composite', false);
print('All Landsat images:',ls);

// Compute a cloud score and mask clouds
ls = ls.map(landsatCloudScore);

// Find and mask out dark outliers
ls = simpleTDOM2(ls,zScoreThresh,shadowSumThresh,dilatePixels);
    
// Run reducer on image collection 
var composite = ls.median();

// Set up the export image.
var toExport = composite.select(
  ['blue','green','red','nir','swir1', 'temp', 'swir2']
  ).clip(studyArea);

// Add the final image to display.
Map.addLayer(toExport.clip(studyArea), vizParamsRescale, 
  "imageall", true);

// Add the final image to display in false color.
Map.addLayer(toExport.clip(studyArea), vizParamsFalseColor, 
  'FalseColorPlot', true);

// Save the study area coordinates to set the region in the export.
var region = studyArea.geometry().bounds().getInfo().coordinates[0];

// Export the image.
Export.image.toDrive({
  image: toExport,
  description: 'LandsatComposite',
  maxPixels: 598075858 ,
  region: studyArea
  });  

/////////////////////////////////////////////////////////////////////////////
function snowScore(img){
      // Compute several indicators of snowyness and take the minimum of them.
      var score = ee.Image(1.0);
      // Snow is reasonably bright in the blue band.
      score = score.min(rescale(img, 'img.blue', [1000, 3000]));
    
      // Snow is reasonably bright in all visible bands.
      score = score.min(rescale(img, 'img.red + img.green + img.blue', [2000, 8000]));
    
      // // Excluded this for snow reasonably bright in all infrared bands.
      // score = score.min(
      //     rescale(img, 'img.nir + img.swir1 + img.swir2', [0.3, 0.8]));
    
      // Snow is reasonably cool in temperature.
      //Changed from [300,290] to [290,275] for AK
      score = score.min(rescale(img, 'img.temp', [3000, 2850]));
      
       // Snow is high in ndsi.
      var ndsi = img.normalizedDifference(['green', 'swir1']);
      ndsi = rescale(ndsi, 'img', [0.5, 0.7]);
      score = score.min(ndsi);
      score = score.clamp(0,1)
      
      score = score.lt(0.01).rename('snowscore');
      img = img.updateMask(img.mask().and(score));
      
      return img.addBands(score);

      }

//////////////////////////////////////////////////////
var shadowSumBands = ee.List(['nir','swir1','swir2']);//Bands for shadow masking

function addShadowSum(img){
    return img.addBands(img.select(shadowSumBands).reduce(ee.Reducer.sum()).select([0],['shadowSum']));
  }
  
function waterScore(img){
      // Compute several indicators of water and take the minimum of them.
      var score = ee.Image(1.0);
      
      
      //Set up some params
      var darkBands = ['green','red','nir','swir2','swir1'];//,'nir','swir1','swir2'];
      var brightBand = 'blue'
      
      //Water tends to be dark
      var sum = img.select(shadowSumBands).reduce(ee.Reducer.sum());
      var sum = rescale(sum,'img',[3500,2000]).clamp(0,1)
      score = score.min(sum)
      
      //It also tends to be relatively bright in the blue band
      var mean = img.select(darkBands).reduce(ee.Reducer.mean());
      var std = img.select(darkBands).reduce(ee.Reducer.stdDev());
      var z = (img.select([brightBand]).subtract(std)).divide(mean)
      z = rescale(z,'img',[0,1]).clamp(0,1)
      score = score.min(z)
      
      // // Water is at or above freezing
      
      score = score.min(rescale(img, 'img.temp', [2730, 2750]));
      
      
      // // Water is nigh in ndsi (aka mndwi)
      var ndsi = img.normalizedDifference(['green', 'swir1']);
      ndsi = rescale(ndsi, 'img', [0.3, 0.8]);
      
      score = score.min(ndsi);
      score = score.clamp(0,1)
      score = score.lt(0.05).rename('waterscore');
      img = img.updateMask(img.mask().and(score));
      
      return img.addBands(score);
      
      }
      
function maskWater(img){
  var ws = waterScore(img)
  return img.mask(img.mask().and(ws.lt(0.05)))
}

// Calculate MNDWI
var mndwi_viz = {min:0, max:0.3, palette:"FFFFFF, 0000FF"};
var mndwi = composite.normalizedDifference(['green', 'swir1']);

var snow = snowScore(composite)
var water_mask = snow.gt(0).select("snowscore")

var maskedComposite = mndwi.updateMask(water_mask)

var out = mndwi
var water_mask1 = maskedComposite.gt(0)
var maskedComposite1 = out.updateMask(water_mask1)

var VIS_OCCURRENCE1 = {min:0, max:0.3, palette: ['blue', 'blue']};
Map.addLayer(maskedComposite1.clip(studyArea), VIS_OCCURRENCE1, "mNDWI");

// Export the image.
Export.image.toDrive({
  image: maskedComposite1.clip(studyArea),
  description: 'LCJ_mNDWI',
  maxPixels: 1e12,
  region: studyArea 
  });
